@@ -12,10 +12,13 @@ from vinca.lib import ansi
 from vinca.lib import filter
 from vinca.generators import generate, GENERATORS_DICT
 
-# TODO: intelligent scrollback for reviewers and editors
-# how? intelligent parameterization
-# TODO: make functions specify the parameters they need and 
-# feed them this explicitly á la FILTER
+# TODO: media cards without vim
+# TODO: better alternative for readline.
+##	tab completion
+##	vim style navigation
+##	up arrows
+##	shortcut for cloze?
+# TODO: convert maps/anatomy to ansi/ascii outlines
 
 TODAY = datetime.date.today()
 DAY = datetime.timedelta(days=1)
@@ -50,8 +53,8 @@ def arg(*names_or_flags, **kwargs):
 	return names_or_flags, kwargs
 CARDS = arg('cards',nargs='*',type=card_type, default=ALL_CARDS)
 CARD = arg('card',type=card_type)
-OPTIONAL_CARD = arg('card', nargs='?', type=card_type)
 OPTIONAL_CARDS = arg('--cards',nargs='*',type=card_type,default=ALL_CARDS)
+OPTIONAL_CARD = arg('card', nargs='?', type=card_type)
 OPTIONAL_CARDS_DEFAULT_NONE = arg('--cards',nargs='*',type=card_type,default=[])
 def subcommand(*subparser_args, parent=subparsers, alias='', name=None):
 	def decorator(func, alias=alias, name=name):
@@ -67,21 +70,21 @@ def subcommand(*subparser_args, parent=subparsers, alias='', name=None):
 	return decorator
 
 @subcommand(arg('-t','--default_tags', nargs='+', default=[]), alias='a')
-def add(args): 
+def add(): 
+	'''Add several cards and browse them.'''
 	ansi.hide_cursor()
 	print(*[f'{key}\t{generator}' for key, generator in GENERATORS_DICT.items()],sep='\n')
 	k = readchar.readchar()
 	if k not in GENERATORS_DICT: return
-	args.generator = GENERATORS_DICT[k]
-	args.cards = generate(args)
+	cards = generate(generator = GENERATORS_DICT[k], scrollback = True)
 	ansi.up_line(len(GENERATORS_DICT) + 2)
 	ansi.clear_to_end()
-	browse(args)
+	browse(card, cards)
 
 @subcommand(OPTIONAL_CARD, OPTIONAL_CARDS, alias='s')
-def statistics(args):
-	if args.card:
-		card = args.card
+def statistics(card, cards, scrollback = False):
+	'''Simple summary statistics. Use `S` for more advanced statistics.'''
+	if card:
 		print(f'\nCard #{card.id}')
 		print(str(card))
 		print(f'Tags: {" ".join(card.tags)}')
@@ -89,62 +92,79 @@ def statistics(args):
 		print('Date\t\tTime\tGrade')
 		print(*[f'{date}\t{time}\t{grade}' for date, time, grade in card.history],sep='\n',end='')
 		lines = 5+len(card.history)
-	elif args.cards:
-		due_cards = filter.filter(args.cards, due_only=True)
-		new_cards = filter.filter(args.cards, new_only=True)
-		print('Total', len(args.cards), sep='\t')
+	elif cards:
+		due_cards = filter.filter(cards, due_only=True)
+		new_cards = filter.filter(cards, new_only=True)
+		print('Total', len(cards), sep='\t')
 		print('Due', len(due_cards), sep='\t')
 		print('New', len(new_cards), sep='\t')
 		lines = 3
-	if args.scrollback:
+	if scrollback:
 		ansi.up_line(lines)
 		ansi.hide_cursor()
 
+@subcommand(OPTIONAL_CARD, OPTIONAL_CARDS, alias='S')
+def advanced_statistics(card, cards):
+	''' Advanced statistics for a card or collection of cards. '''
+	if card:
+		raise NotImplementedError
+	elif cards:
+		raise NotImplementedError
+
 @subcommand(CARD,alias='e')
-def edit(args):
+def edit(card):
 	''' edit a single card '''
-	args.card.edit()
+	card.edit()
 
 @subcommand(CARD,alias='E')
-def edit_metadata(args):
+def edit_metadata(card):
 	''' edit the metadata of a card '''
-	args.card.edit_metadata()
+	card.edit_metadata()
 
 @subcommand(OPTIONAL_CARD, OPTIONAL_CARDS_DEFAULT_NONE, alias='x')
-def delete(args):
-	if args.card:
-		args.card.deleted = not args.card.deleted
-	elif args.cards:
-		for card in args.cards:
+def delete(card, cards):
+	''' Delete (or undelete) a card. These cards will continue to be stored. Use purge to permanently delete these cards.'''
+	if card:
+		card.deleted = not card.deleted
+	elif cards:
+		for card in cards:
 			card.deleted = not card.deleted
 
 @subcommand(CARDS, alias='r')
-def review(args):
-	if args.card:
-		args.card.review()
-		args.card.schedule()
-	elif args.cards:
-		args.cards = filter.filter(args.cards, due_only = True)
-		if not args.cards:
+def review(card, cards):
+	''' Review a card or collection of cards. Grading is as follows:
+	1 hard
+	2 again
+	3 good
+	4 easy. '''
+	if card:
+		card.review()
+		card.schedule()
+	elif cards:
+		cards = filter.filter(cards, due_only = True)
+		if not cards:
 			print('No cards due.')
 			return
-		browse(args, reviewing = True)
+		browse(None, cards, reviewing = True)
 
 @subcommand(arg('tag'), OPTIONAL_CARD, OPTIONAL_CARDS_DEFAULT_NONE)
-def add_tag(args):
-	for card in [args.card] if args.card else args.cards:
-		card.tags += [args.tag]
+def add_tag(card, cards, tag):
+	''' Add a single tag to a collection of cards. '''
+	for card in [card] if card else cards:
+		card.tags += [tag]
 
 @subcommand(arg('tag'), OPTIONAL_CARD, OPTIONAL_CARDS_DEFAULT_NONE)
-def remove_tag(args):
-	for card in [args.card] if args.card else args.cards:
-		if args.tag in card.tags:
-			card.tags.remove(args.tag)
+def remove_tag(card, cards, tag):
+	''' Remove a single tag from a collection of cards. '''
+	for card in [card] if card else cards:
+		if tag in card.tags:
+			card.tags.remove(tag)
 		# TODO do this with set removal
 		card.save_metadata()
 
 @subcommand(OPTIONAL_CARD, OPTIONAL_CARDS_DEFAULT_NONE, alias='t')
-def edit_tags(args):
+def edit_tags(card, cards, scrollback = False):
+	# TODO rewrite completion
 	def complete(text, state):
 		for tag in ALL_TAGS:
 			if tag.startswith(text):
@@ -155,40 +175,36 @@ def edit_tags(args):
 	rl.parse_and_bind('tab: complete')
 	rl.set_completer(complete)
 
-	if args.card:
+	if card:
 		pass #TODO lineedit tags with tab complete
-		tags = ' '.join(args.card.tags)
+		tags = ' '.join(card.tags)
 		rl.set_startup_hook(lambda: rl.insert_text(tags))
-		args.card.tags = input('tags: ').split()
+		card.tags = input('tags: ').split()
 		lines = 1
 		rl.set_startup_hook()
-	elif args.cards:
+	elif cards:
 		tags_add = input('tags to add: ').split()
 		tags_remove = input('tags to remove: ').split()
 		lines = 2
 		for tag in tags_add:
-			args.tag = tag
-			add_tag(args)
+			add_tag(card = card, cards = cards, tag = tag)
 		for tag in tags_remove:
-			args.tag = tag
-			remove_tag(args)
+			remove_tag(cards = cards, tag = tag)
 	
 	rl.set_completer()
 
-	if args.scrollback:
+	if scrollback:
 		ansi.up_line(lines)
 
 
 CMD_DICT = {'r': review, 'R': review,
-	    's': statistics, 'S': statistics,
+	    's': statistics, 'S': advanced_statistics,
 	    'x': delete, 'X': delete,
 	    'e': edit, 'E': edit_metadata,
 	    't': edit_tags, 'T': edit_tags}
 @subcommand(arg('--cards',nargs='*',type=card_type,default=filter.filter(ALL_CARDS)), arg('-t','--default_tags', nargs='+', default=[]), alias='b')
-def browse(args, reviewing = False):
+def browse(card, cards, reviewing = False):
 	''' Use J and K to move up and down. In general a lowercase letter will act on the selected card while a capital letter will act upon all cards. Use the following hotkeys in browse mode: r, x, s, e, t'''
-	args.scrollback = True
-	cards = args.cards
 	if not cards:
 		print('no cards')
 		return
@@ -208,7 +224,7 @@ def browse(args, reviewing = False):
 			print(f'{sel+1} of {len(cards)}')
 			ansi.reset()
 		for i, card in enumerate(cards[frame:frame+FRAME_WIDTH], start=frame):
-			if card.due_as_of(args.review_date):
+			if card.due_as_of(TODAY):
 				ansi.bold()
 				ansi.blue()
 			if card.deleted:
@@ -244,8 +260,8 @@ def browse(args, reviewing = False):
 				quit()
 		if k in CMD_DICT:
 			ansi.show_cursor()
-			args.card = cards[sel] if k in ('x','s','r','R','e','E', 't') else None
-			CMD_DICT[k](args)
+			card = cards[sel] if k in ('x','s','r','R','e','E', 't') else None
+			CMD_DICT[k](card, cards)
 			ansi.hide_cursor()
 			reviewing = (k == 'R' and cards[sel].last_grade != 'exit')
 			if reviewing and sel == N - 1:
@@ -254,8 +270,7 @@ def browse(args, reviewing = False):
 				sel += 1 
 		if k in GENERATORS_DICT:
 			ansi.show_cursor()
-			args.generator = GENERATORS_DICT[k]
-			new_cards = generate(args)
+			new_cards = generate(generator = GENERATORS_DICT[k])
 			cards[sel:sel] = new_cards
 			N = len(cards)
 			STATUS_BAR = N > FRAME_WIDTH
@@ -284,26 +299,27 @@ def browse(args, reviewing = False):
 	arg('--due_date_min',type=date_type, metavar='DATE'),
 	arg('--due_date_max',type=date_type, metavar='DATE'),
 	arg('--due_only',action='store_true'),
-	arg('--not_due_only',action='store_true'),
 	arg('--editor', type=str),
 	arg('--reviewer', type=str),
 	arg('--scheduler', type=str),
 	arg('--deleted_only',action='store_true'),
-	arg('--show_deleted',action='store_true'),
 	arg('--new_only',action='store_true'),
-	arg('--not_new_only',action='store_true'),
 	alias='f', name='filter')
-def display_filter(args):
+def display_filter(pattern, invert, id_only, cards,
+		   tags_include, tags_exclude, create_date_min, create_date_max,
+		   seen_date_min, seen_date_max, due_date_min, due_date_max,
+		   due_only, deleted_only, new_only,
+		   editor, reviewer, scheduler):
 	''' Filter a set of cards using a variety of parameters.
 	It is often helpful to pipe this command to another one.'''
 
-	# get filter parameters as a list of strings
-	filter_kwargs = inspect.getfullargspec(filter.filter).args[1:]
-	# check that args specifies these parameters
-	assert all([hasattr(args, param) for param in filter_kwargs])
-	matches = filter.filter(args.cards,
-		# feed the keyword args editor=args.editor, due=args.due, 
-		**{param : getattr(args, param) for param in filter_kwargs})
+	options = locals()
+
+	# get a list of the filter parameters
+	parameters = inspect.signature(filter.filter).parameters
+	# pass the list of options this function received on to the filtering function
+	matches = filter.filter(**{p : options[p] for p in parameters})
+	
 	if PIPE_OUT:
 		print(*[card.id for card in matches], sep='\n', end='')
 		exit()
@@ -321,7 +337,7 @@ def display_filter(args):
 		ansi.reset()
 	
 @subcommand()
-def purge(args):
+def purge():
 	''' Permanently delete all cards marked for deletion. '''
 	deleted_cards = filter.filter(ALL_CARDS, deleted_only=True)
 	if not deleted_cards:
@@ -333,18 +349,20 @@ def purge(args):
 			rmtree(card.path)
 
 @subcommand(arg('backup_path',type=Path), CARDS, name='export')
-def exp(args):
-	for card in args.cards:
-		copytree(card.path, args.backup_path / str(card.id))
+def exp(cards, backup_path):
+	''' Save the collection of cards to a new folder. '''
+	for card in cards:
+		copytree(card.path, backup_path / str(card.id))
 
 @subcommand(arg('import_path',type=Path), arg('-o','--overwrite', action='store_true'), name='import')
-def imp(args):
-	if args.overwrite:
+def imp(import_path, overwrite = False):
+	''' Import a collection of cards. Specify the folder in which these cards are located. '''
+	if overwrite:
 		rmtree(cards_path)
-		copytree(args.import_path, cards_path)
+		copytree(import_path, cards_path)
 		return
 	old_ids = [card.id for card in ALL_CARDS]
-	for new_id,card_path in enumerate(args.import_path.iterdir(), max(old_ids, default=1) + 1):
+	for new_id,card_path in enumerate(import_path.iterdir(), max(old_ids, default=1) + 1):
 		copytree(card_path, cards_path / str(new_id))
 
 
@@ -356,7 +374,7 @@ for alias, generator in GENERATORS_DICT.items():
 
 # parse the command line arguments
 parser.add_argument('-V','--version',action='version',version=str(VERSION))
-parser.set_defaults(cards = [], func = lambda args: parser.print_help(), review_date = TODAY, scrollback = False, card = None)
+parser.set_defaults(cards = [], func = lambda: parser.print_help(), card = None)
 # simplifiy the help message
 parser._action_groups[0].title = 'commands'
 parser._action_groups.pop()
@@ -366,4 +384,11 @@ if PIPE_IN:
 	ids = [int(line.strip().split()[0]) for line in sys.stdin]
 	args.cards = [Card(id) for id in ids]
 	sys.stdin = open('/dev/tty')  
-args.func(args)
+
+# from args pass to the function the parameters which it requires.
+# get filter parameters as a list of strings
+parameters = inspect.signature(args.func).parameters.values()
+specified_parameters = [p.name for p in parameters if hasattr(args, p.name)]
+required_parameters = [p.name for p in parameters if p.default is inspect._empty]
+assert all(p in specified_parameters for p in required_parameters)
+args.func(**{p : getattr(args, p) for p in specified_parameters})
