@@ -1,6 +1,7 @@
 from vinca.lib import ansi
 from vinca.lib.terminal import COLUMNS
 from vinca.lib import readchar
+from vinca.lib.readchar import ESC, BACK
 import string
 
 def input(prompt = '', default_text = '', completions = None):
@@ -16,14 +17,12 @@ def input(prompt = '', default_text = '', completions = None):
 		c = readchar.readkey()
 		if c in ('\n','\r', ESC):
 			break
-		elif c in ('\b', BACKSPACE):
+		elif c in ('\b',BACK):
 			text = text[:pos - 1] + text[pos:]
 			pos -= 1
-		elif c == DEL:
-			text = text[:pos] + text[pos + 1:]
 		elif c in ('\t'):
 			pass
-		elif c in (UP, DOWN, LEFT, RIGHT):
+		elif c in []:
 			if c == LEFT:
 				pos -= 1
 			if c == RIGHT:
