@@ -1,6 +1,7 @@
 ESC = chr(27)
 ESCC = ESC + '['
-DEL = ESCC + '3~'
+DEL = DELETE = ESCC + '3~'
+BACK = BACKSPACE = chr(127)
 UP, DOWN, LEFT, RIGHT = ESCC + chr(65), ESCC + chr(66), ESCC + chr(67) ,ESCC + chr(68)
 KEYS = (DEL, UP, DOWN, LEFT, RIGHT)
 
