@@ -2,11 +2,10 @@ import importlib
 
 def schedule(card):
 
-	if card.last_grade == -1:
-		card.delete()
-		return 0
+	if card.last_grade == 'delete':
+		card.deleted = True
 
 	# import the specific scheduler module
 	m = importlib.import_module('.' + card.scheduler, package = 'vinca.schedulers')
 	# invoke the specific scheduler
-	return m.schedule(card)
+	m.schedule(card)
